# ***************************************************************************
# Copyright 2014-2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import argparse
import errno
import getpass
import json
import logging
from logging import FileHandler
from logging.handlers import RotatingFileHandler
from logging.handlers import SysLogHandler
import nsx_logging
import nsx_logging.config
import os
import re
import shutil
import socket
from stat import S_IMODE
from stat import S_IRGRP
from stat import S_IRUSR
from stat import S_IWGRP
from stat import S_IWUSR
import subprocess
import sys
import threading
import time
import traceback
import yaml

from cli.descriptors.descriptor_factory import DescriptorFactory
from cli.server.cli_command_service import CommandService
from cli.utils.cache_utils import CentralCLICacheThread
from cli.utils import rest_utils
from cli.utils.central_cli_utils import is_central_cli_command
from cli.utils.command_utils import invoke_command
from cli.utils.constants import CENTRAL_CLI_STR
from cli.utils.constants import CLI_INTERRUPTED_WHILE_LOAD
from cli.utils.constants import CMD_EXECUTED
from cli.utils.constants import CMD_EXECUTED_REQUESTED_EXIT
from cli.utils.constants import CMD_EXECUTED_WITH_ERROR_RESULT
from cli.utils.constants import CMD_NOT_EXECUTED
from cli.utils.constants import CMD_SESSION_TIMEOUT
from cli.utils.constants import CMD_UNEXPECTED_ERROR
from cli.utils.constants import CMD_UNEXPECTED_EXCEPTION
from cli.utils.constants import CMD_UNSUPPORTED_MODE
from cli.utils.constants import DEVICE_LOG_PATH
from cli.utils.constants import MODE_ENABLE
from cli.utils.constants import NODE_TYPE_CLOUD_SERVICE_MANAGER
from cli.utils.constants import NODE_TYPE_NSXAAS_DASHBOARD
from cli.utils.constants import NODE_TYPE_CONTAINER_KUBE_PROXY
from cli.utils.constants import NODE_TYPE_CONTAINER_NCP
from cli.utils.constants import NODE_TYPE_CONTAINER_NODE_AGENT
from cli.utils.constants import NODE_TYPE_CONTROLLER
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_PUBLIC_CLOUD
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_PACE
from cli.utils.constants import NODE_TYPE_GLOBAL_MANAGER
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.constants import NODE_TYPE_KEYMANAGER
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NODE_TYPE_POLICY
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.constants import PERMISSION_CRUD
from cli.utils.constants import PERMISSION_READ
from cli.utils.error_utils import render_error
from cli.utils.logging_utils import clear_log_directory
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import get_node_type_string
from cli.utils.node_utils import is_appliance
from cli.utils.node_utils import is_container
from cli.utils.node_utils import is_windows
from cli.utils.node_utils import should_enable_central_cli
from cli.utils.prompt_utils import CliPrompt
from cli.utils.error_code import ErrorCode as ec
from compat23 import compat23_httplib
if not is_windows():
    import fcntl
    import grp
    from cli.utils.pcap_utils import stop_all_packet_capture_to_file
_NODE_TYPE = get_node_type()


lg = None


LOG_FILE_MODE = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP
LOG_FORMAT_FILE = "%(asctime)s %(process)d %(name)s %(levelname)s %(message)s"

COMMENT_CHARS = ["!"]
logging.raiseExceptions = False

_LOG_NAME = "nsxcli.log"
_LOG_LOCK_FILENAME = ".nsxcli_lock_file"
_TMP_LOG_FILE = os.path.join("/", "tmp", _LOG_NAME)
_LOG_MAX_BYTES = 2 * 1024 * 1024
_LOG_BACKUP_COUNT = 5
_NSXCLI_GRP_NAME = "nsxcli"
_NSXCLI_GRP_ID = 10001
_READONLY_UIDS = [10002]
_UPGRADE_FLAG = "/image/current_bundle_path"
_UPGRADE_MSG = "/etc/upgrade.msg"
_AUTO_EDGE_BOOTUP_ERROR = "/config/vmware/edge/bootup.err"
_EDGE_BOOTUP_ERROR = "/config/vmware/edge/crypto_integrity.err"
_SECONDARY_PARTITION_ABSENT = "/image/.ua_secondary_partition_absent"
_SW_INTEGRITY_CHECK_FAILED = "/image/.sw_integrity_check_failed"

# The following dictionary maps a node_type as returned by get_node_type()
# to a top level yml file found in the cli/commands directory. The top level
# yml file is used to specify all commands available for a given node type.
# TODO: in a future patch, we may want to rename top level yml files to
# NODE_TYPE.yml at which point _NODE_TYPE_YML_MAP is no longer needed.
dir_path = os.path.dirname(os.path.realpath(__file__))
_TOP_LEVEL_YML_DIR = os.path.join(dir_path, os.path.pardir, 'commands')
_NODE_TYPE_YML_MAP = {
    NODE_TYPE_CLOUD_SERVICE_MANAGER: "csm.yml",
    NODE_TYPE_NSXAAS_DASHBOARD: "nsxaas-dashboard.yml",
    NODE_TYPE_CONTAINER_KUBE_PROXY: "container_kube_proxy.yml",
    NODE_TYPE_CONTAINER_NCP: "container_ncp.yml",
    NODE_TYPE_CONTAINER_NODE_AGENT: "container_node_agent.yml",
    NODE_TYPE_CONTROLLER: "controller.yml",
    NODE_TYPE_EDGE: "edge.yml",
    NODE_TYPE_AUTONOMOUS_EDGE: "autonomous_edge.yml",
    NODE_TYPE_HOST_CENTOSKVM: "host_kvm.yml",
    NODE_TYPE_HOST_SLESKVM: "host_kvm.yml",
    NODE_TYPE_HOST_ESX: "host_esx.yml",
    NODE_TYPE_HOST_PUBLIC_CLOUD: "host_public_cloud.yml",
    NODE_TYPE_HOST_RHELKVM: "host_kvm.yml",
    NODE_TYPE_HOST_UBUNTUKVM: "host_kvm.yml",
    NODE_TYPE_HOST_WIN: "host_kvm.yml",
    NODE_TYPE_GLOBAL_MANAGER: "global_manager.yml",
    NODE_TYPE_BAREMETAL_UBUNTU_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_RHEL_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_CENTOS_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_SLES_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_OEL_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_WINDOWS_SERVER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_RHEL_CONTAINER: "host_kvm.yml",
    NODE_TYPE_BAREMETAL_CENTOS_CONTAINER: "host_kvm.yml",
    NODE_TYPE_KEYMANAGER: "keymanager.yml",
    NODE_TYPE_MANAGER: "manager.yml",
    NODE_TYPE_POLICY: "policy.yml",
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY: "publiccloudgateway.yml"
}

# Disallow following users for interactive invocation of the CLI
DISALLOWED_TTY_USERS = ('root', )


def _get_supported_users():
    """Return string of supported users for interactive CLI"""
    import pwd # Late import as it's only availbale on all Unix versions
    try:
        supported_users = ' or '.join([pwd.getpwuid(i)[0] for i in (10000, 10002)])
    except:
        supported_users = 'admin or audit'
    return supported_users


def _get_config():
    """Return dictionary of user specified command line arguments."""
    parser = argparse.ArgumentParser(description="NSXCLI", prog="nsxcli")
    parser.add_argument("-c", "--cmd", nargs="*")
    parser.add_argument("--log_file_path")
    parser.add_argument("--log_level", type=int)
    parser.add_argument("--raw_output", action="store_true")
    args = parser.parse_args()
    return vars(args)


class MultiprocessRotatingFileHandler(RotatingFileHandler):
    """Handler for use by multiple processes to log to same file.

    This handler subclasses RotatingFileHandler to support writes
    by multiple processes to the same log file. The default python
    logging libraries only support single process writes to file.
    """
    def __init__(self, filename, lock_filename, node_type, mode='a', maxBytes=0,
                 backupCount=0, encoding=None):
        self._lock_filename = lock_filename
        self._node_type = node_type
        super(MultiprocessRotatingFileHandler, self).__init__(
            filename, mode, maxBytes, backupCount, encoding, delay=1)

    def emit(self, record):
        # Note, exceptions are ignored when emitting a log record because
        # logging the exception will likely raise another exception.
        locked = False
        lock_fd = None
        try:
            lock_fd = open(self._lock_filename, "w")
            attempts = 100  # Should never have to retry this many times
            while attempts:
                try:
                    fcntl.flock(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                    locked = True
                    break
                except IOError:
                    # The exception is likely EACCES or EAGAIN because
                    # the lock file is locked.
                    time.sleep(0.05)
                    attempts -= 1
            if locked:
                if self.stream is None:
                    self.stream = self._open()
                if self._node_type == NODE_TYPE_HOST_ESX:
                    self._esx_pre_emit(record)
                super(MultiprocessRotatingFileHandler, self).emit(record)
        except:
            pass
        finally:
            try:
                if self.stream:
                    self.close()
            except:
                pass
            if lock_fd is not None:
                if locked:
                    try:
                        fcntl.flock(lock_fd, fcntl.LOCK_UN)
                    except Exception:
                        pass
                try:
                    lock_fd.close()
                except Exception:
                    pass

    def _open(self):
        """Overload RotatingFileHandler._open to set file permissions."""
        # On ESX, log files under /var/log should be a symlink to
        # /scratch/log. Note that the /scratch/log directory may not be
        # available. If this is the case just create the log file under
        # /var/log for now and move it on rotate or next open.
        if self._node_type == NODE_TYPE_HOST_ESX:
            mode = S_IRUSR | S_IWUSR
            scratch_path = self._esx_log_path(
                os.path.basename(self.baseFilename))
            if not os.path.isfile(self.baseFilename):
                try:
                    open(scratch_path, self.mode).close()
                    os.chmod(scratch_path, mode)
                    os.symlink(scratch_path, self.baseFilename)
                except:
                    # If we get here, creating the link from /var/log to
                    # /scratch/log failed, just open file under /var/log
                    pass
            # If /var/log file is currently not a link, attempt move to
            # /scratch/log
            elif not os.path.islink(self.baseFilename):
                try:
                    shutil.move(self.baseFilename, scratch_path)
                    os.symlink(scratch_path, self.baseFilename)
                except:
                    # Move failed, continue using file under /var/log
                    pass
        else:
            mode = LOG_FILE_MODE
        stream = super(MultiprocessRotatingFileHandler, self)._open()
        if is_appliance():
            # When log rotation occurs, the mode may be different from what it
            # should be, LOG_FILE_MODE. Furthermore, the group ownership is
            # assigned to the current user's group instead of what it should
            # be, _NSXCLI_GRP_NAME.  Check mode and group ownership and adjust
            # as needed.
            filestat = os.stat(self.baseFilename)
            if filestat.st_gid != _NSXCLI_GRP_ID:
                os.chown(self.baseFilename, -1, _NSXCLI_GRP_ID)
            if S_IMODE(filestat.st_mode) != LOG_FILE_MODE:
                os.chmod(self.baseFilename, LOG_FILE_MODE)
        else:
            os.chmod(self.baseFilename, mode)
        return stream

    def _esx_log_path(self, basename):
        return os.path.join("/", "scratch", "log", basename)

    def _esx_move_var_log_path(self, basename):
        src = os.path.join("/", "var", "log", basename)
        if os.path.isfile(src) and not os.path.islink(src):
            dst = self._esx_log_path(basename)
            try:
                os.remove(dst)
            except:
                pass  # EAFP
            try:
                shutil.move(src, dst)
            except:
                pass  # Ignore for now, retry on next rotate

    def _esx_pre_emit(self, record):
        # Before emitting, perform rotation (if needed) and move any rotated
        # files that are still in /var/log to /scratch/log
        if self.shouldRollover(record):
            scratch_path = self._esx_log_path(
                os.path.basename(self.baseFilename))
            scratch_dir = os.path.split(scratch_path)[0]
            baseFilename = self.baseFilename
            try:
                if os.path.isfile(scratch_path):
                    self.baseFilename = scratch_path
                if os.path.isdir(scratch_dir):
                    # First move all the rotated nsxcli.log.# files
                    name = os.path.basename(baseFilename)
                    for i in range(1, self.backupCount + 1):
                        self._esx_move_var_log_path("%s.%s" % (name, i))
                    # Now move the nsxcli.log file
                    if self.stream:
                        self.stream.close()
                        self.stream = None
                        self._esx_move_var_log_path(name)
                    self.baseFilename = scratch_path
                # The Python doRollover method rotates self.baseFilename,
                # so the above code should have set it to the scratch_path
                # already
                self.doRollover()
            finally:
                self.baseFilename = baseFilename


class SyslogNonAuditFilter(logging.Filter):
    """Syslog log filter to allow non-audit entries"""
    def filter(self, record):
        return not record.name.startswith("cli.audit")


class SyslogAuditFilter(logging.Filter):
    """Syslog log filter to only allow audit entries"""
    def filter(self, record):
        return record.name.startswith("cli.audit")


class SyslogRecordFormatter(logging.Formatter):
    """Syslog formatter to populate hostname, APP-NAME, PROCID, MSGID
       as well as other NSX structured data
    """
    # Message format that complies with RFC5424, uses the following values:
    # - Headers:
    #   - PRI:
    #   - TIMESTAMP:
    #   - HOSTNAME:
    #   - APP-NAME:     NSX
    #   - PROCID:       NSX CLI pid
    #   - MSGID:        SYSTEM
    # - IANA defined structured data:
    #   - enterpriseId: 6876 (VMware)
    # - NSX defined structured data:
    #   - comp:         nsx-manager
    #   - subcomp:      node-cli
    #   - username:     admin (currently the only interactive CLI user is admin)
    component = "nsx"
    if is_appliance():
        component = (
            "%s-%s" %
            (component, get_node_type_string().replace(" ", "").lower()))
    LOG_FORMAT = (
        "1 %%(asctime)s %%(hostname)s NSX %%(process)d SYSTEM "
        "[nsx@6876 comp=\"%s\" subcomp=\"node-cli\" username=\"%s\" "
        "level=\"%%(levelname)s\"%%(tags)s%s] %%(message)s")
    DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
    converter = time.gmtime  # Overload logging.Formatter.converter

    def __init__(self, custom_struct_data=None):
        data = "" if not custom_struct_data else " %s" % custom_struct_data
        self._log_format = (
            self.LOG_FORMAT % (self.component, getpass.getuser(), data))
        super(SyslogRecordFormatter, self).__init__(
                            self._log_format, self.DATE_FORMAT)

    def format(self, record):
        # Add custom structured data to syslog
        sd_tags = ""
        sd_map = record.__dict__.get("tags")  # record may not have sd tags
        if sd_map:
            sd_items = ['%s="%s"' % (k, v) for k, v in list(sd_map.items())]
            sd_tags = " %s" % " ".join(sd_items)
        record.__dict__["tags"] = sd_tags

        # %(hostname)s is not a standard format defined by python's LogRecord
        # class. We populate it here so it can be set in LOG_FORMAT.
        record.__dict__["hostname"] = socket.gethostname()
        return super(SyslogRecordFormatter, self).format(record)


def _get_configured_scratch_path():
    """Returns the configured path for scratch dir in ESX or,
     default (/scratch/log) if scratch path config is missing."""
    import vmsyslog.cli
    esxcfg = vmsyslog.cli.SyslogCliFactory('esxcfg')
    return esxcfg.get_logdir()


def _get_sys_logger(level):
    """Returns syslog logger"""
    logger = logging.getLogger()
    logger.setLevel(level)
    handler = SysLogHandler(address = DEVICE_LOG_PATH)
    formatter = logging.Formatter("nsx-cli[%(process)d]: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def _config_logger(log_file, log_name, level=nsx_logging.INFO):
    """Configure and return the cli and audit logger objects.

    Note: The audit logger object is typically pointing to syslog and uses
    facility LOCAL7. However, if syslog is unavailable or we are running on
    a host, the audit logger is the same as the cli logger.
    """
    # get default configurations
    cli_path = os.path.realpath(
                    os.path.join(os.path.dirname(__file__), ".."))
    config_path = os.path.join(cli_path, "utils", "logconfig.yml")
    with open(config_path) as stream:
        try:
            # log_config is a logging config dict
            # since each environment has different requirement for logging, we
            # will slightly modify the config dict based on platform running on.
            log_config = yaml.load(stream, Loader=yaml.Loader)
        except:
            nsx_logging.error("logging configuration file error",
                    comp="nsx-manager", subcomp="cli",
                    error_code=ec.INVALID_CONFIG)
            raise
    log_config["loggers"][""]["level"] = level
    lock_filename = os.path.join("/", "var", "tmp", _LOG_LOCK_FILENAME)

    if log_file:
        # In Bumblebee, log_file is specified when NSX CLI is invoked on hosts
        # because we can't assume:
        # - hosts which are managed by customers have syslog configured.
        # - the standard /var/log directory exists and is used for logging.
        # so we will not log to syslog, and we won't do multi-process logging on
        # hosts.
        log_config["loggers"][""]["handlers"].remove("syslog")
        del log_config["handlers"]["syslog"]
        del log_config["handlers"]["file"]["lockname"]
        log_config["handlers"]["file"]["filename"] = log_file
    elif _NODE_TYPE in [NODE_TYPE_HOST_SLESKVM,
                        NODE_TYPE_HOST_CENTOSKVM,
                        NODE_TYPE_HOST_ESX,
                        NODE_TYPE_HOST_RHELKVM,
                        NODE_TYPE_HOST_UBUNTUKVM,
                        NODE_TYPE_HOST_PUBLIC_CLOUD,
                        NODE_TYPE_BAREMETAL_UBUNTU_SERVER,
                        NODE_TYPE_BAREMETAL_RHEL_SERVER,
                        NODE_TYPE_BAREMETAL_CENTOS_SERVER,
                        NODE_TYPE_BAREMETAL_SLES_SERVER,
                        NODE_TYPE_BAREMETAL_OEL_SERVER,
                        NODE_TYPE_BAREMETAL_WINDOWS_SERVER,
                        NODE_TYPE_BAREMETAL_RHEL_CONTAINER,
                        NODE_TYPE_BAREMETAL_CENTOS_CONTAINER]:
        log_dir = os.path.join("/", "var", "log")
        if _NODE_TYPE is NODE_TYPE_HOST_ESX:
            sys_logger = _get_sys_logger(level)

            # On ESXi, /var/log contains symlinks to actual log files stored
            # under scratch dir
            try:
                scratch_dir = _get_configured_scratch_path()
            except Exception as ex:
                sys_logger.error("Exception while getting scratch dir "
                                 "from localcli" + ": [%s] %s" %
                                 (ex.__class__.__name__, ex))
                scratch_dir = None

            # Incase of missing scratch dir path in file system,
            # fallback on creating log & lock file directly in /var/log
            _dir = scratch_dir if scratch_dir and os.path.exists(scratch_dir)\
               else log_dir

            local_log_file = os.path.join(_dir, _LOG_NAME)
            try:
                # If log file doesn't exists,
                # touch create log file on scratch path
                if not os.path.exists(local_log_file):
                    open(local_log_file, "a").close()
                    os.chmod(local_log_file, S_IRUSR | S_IWUSR)

                var_log = os.path.join(log_dir, _LOG_NAME)
                # Link the logs created on scratch path to /var/log
                # when symlink/dir not exists
                if not os.path.exists(var_log):
                    # Unlink existing symlink if present
                    if os.path.islink(var_log):
                        os.unlink(var_log)
                    os.symlink(local_log_file, var_log)
            except Exception as ex:
                # Exit from the system if unable to create log file
                sys_logger.error("Exception while initializing " +
                                  "log file: [%s] %s" %
                                  (ex.__class__.__name__, ex))
                sys.exit(-1)
        else:
            local_log_file = os.path.join(log_dir, _LOG_NAME)
        # We can't be certain whether syslog is enabled on hosts,
        # so don't log to syslog
        del log_config["handlers"]["syslog"]
        log_config["loggers"][""]["handlers"].remove("syslog")
        log_config["handlers"]["file"]["filename"] = local_log_file
        log_config["handlers"]["file"]["lockname"] = lock_filename
    elif _NODE_TYPE in [NODE_TYPE_CONTAINER_KUBE_PROXY,
                        NODE_TYPE_CONTAINER_NCP,
                        NODE_TYPE_CONTAINER_NODE_AGENT]:
        log_dir = os.path.join("/", "var", "log", "nsx-cli")
        local_log_file = os.path.join(log_dir, _LOG_NAME)
        del log_config["handlers"]["syslog"]
        log_config["loggers"][""]["handlers"].remove("syslog")
        log_config["handlers"]["file"]["filename"] = local_log_file
        log_config["handlers"]["file"]["lockname"] = lock_filename
    else:
        # We are on an NSX appliance, create a rotating log file
        # and send log entires to syslog if it's available
        try:
            # Some NSX services invoke the NSX CLI via a system call/subprocess
            # even though they're not supposed to. To prevent file permission
            # issues when writing log entries, we only allow interactive users
            # who are in group "nsxcli" to write to /var/log/nsx-cli.
            try:
                nsxcli_members = grp.getgrnam(_NSXCLI_GRP_NAME).gr_mem
                use_var_log = getpass.getuser() in nsxcli_members
            except Exception:
                use_var_log = False
            log_dir = os.path.join("/", "var", "log", "nsx-cli")
            local_log_file = os.path.join(log_dir, _LOG_NAME)
            local_log_file = local_log_file if use_var_log else _TMP_LOG_FILE
        except Exception:
            pass
        log_config["handlers"]["file"]["filename"] = local_log_file
        log_config["handlers"]["file"]["lockname"] = lock_filename
        log_config["handlers"]["file"]["grpname"] = "nsxcli"

    log_config["formatters"]["rfc5424"]["comp"] = nsx_logging._get_nsx_node_type()
    try:
        nsx_logging.config.dictConfig(log_config)
    except:
        lg.exception("Error configuring syslog")

def _print_bootup_error_file(lg, fname):
    """ Print content in given fname to stdout if it exists"""
    if not os.path.isfile(fname):
        return
    try:
        with open(fname) as fd:
            bootup_error = fd.read()
            if bootup_error:
                sys.stdout.write(bootup_error)
                sys.stdout.write("\n")
    except IOError as ex:
        if ex.errno != errno.ENOENT:
            lg.exception("IOError reading bootup err file",
                         error_code=ec.IO_ERROR)
    except Exception:
        lg.exception("Exception reading bootup err file",
                     error_code=ec.INTERNAL_ERROR)

def _print_secondary_disk_error(lg):
    """Print error to stdout if services are not up because of absent secondary disk"""
    if NODE_TYPE_MANAGER in _NODE_TYPE:
        if os.path.isfile(_SECONDARY_PARTITION_ABSENT):
           sys.stdout.write("Services are down because all disks are not present. "
                            "Please refer troubleshooting guide.\n")
           lg.info("Display secondary partition error")

def _print_sw_integrity_check_error(lg):
    """Print error to stdout if services are not up because of software integrity check failure"""
    if os.path.isfile(_SW_INTEGRITY_CHECK_FAILED):
       sys.stdout.write("Services are down because software integrity check failed. "
                        "Please refer troubleshooting guide.\n")
       lg.info("Display software integrity check error")

def _print_bootup_error(lg):
    """Print content of /config/vmware/edge/bootup.err to stdout if it exists"""
    if _NODE_TYPE == NODE_TYPE_AUTONOMOUS_EDGE:
        _print_bootup_error_file(lg, _AUTO_EDGE_BOOTUP_ERROR)

    """Print content of /config/vmware/edge/crypto_integrity.err to stdout if it exists"""
    if _NODE_TYPE in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
        _print_bootup_error_file(lg, _EDGE_BOOTUP_ERROR)

def _print_motd_tail(lg):
    """Print contents of /etc/motd.tail to stdout if it exists."""
    if is_windows():
        return
    try:
        with open("/etc/motd.tail") as fd:
            motd = fd.read()
            if motd:
                sys.stdout.write(motd)
                sys.stdout.write("\n")
    except IOError as ex:
        if ex.errno != errno.ENOENT:
            lg.exception("IOError reading motd.tail file",
                         error_code=ec.IO_ERROR)
    except Exception:
        lg.exception("Exception reading motd.tail file",
                     error_code=ec.INTERNAL_ERROR)


def _print_upgrade_status(lg):
    """Print upgrade status to stdout if node upgrade is in progress."""
    if is_windows():
        return
    try:
        if os.path.isfile(_UPGRADE_FLAG):
           with open(_UPGRADE_MSG) as fd:
               um = fd.read()
               if um:
                   sys.stdout.write(um)
                   sys.stdout.write("\n")
    except IOError as ex:
        if ex.errno != errno.ENOENT:
            lg.exception("IOError reading message file",
                         error_code=ec.IO_ERROR)
    except Exception:
        lg.exception("Exception reading message file",
                     error_code=ec.INTERNAL_ERROR)


def _get_nsx_version(lg):
    status, obj, err, _ = rest_utils.local_node_request("GET", "/node")
    if status != compat23_httplib.OK or obj is None:
        lg.error("Error getting node version, status: %s, obj: %s, err: %s",
                 status, obj, err, error_code=ec.OPERATION_FAILED)
        return None
    return obj.get("node_version")


def _get_user_permissions():
    """Return permission granted to user.

    Return:
        One of the following permission strings:
        crud, execute, read, none.
    """
    # In Dropkick, NSX CLI permissions are system defined:
    # - admin: (invoked on nsx appliances) crud
    # - audit: (invoked on nsx appliances) read
    # - root (invoked on hosts): crud
    # Note, some NSX processes invoke the NSX CLI (even though the NSX CLI is
    # developed for use for human users only). For these processes, we will
    # assign crud permissions for now to maintain backward compatibility.
    if not is_appliance():
        return PERMISSION_CRUD
    return PERMISSION_READ if os.getuid() in _READONLY_UIDS else PERMISSION_CRUD


def admin_passwd_required():
    if is_windows():
        return False
    passwd_reqd = True  # Assume true on error
    cmd = ['chage', '-l', getpass.getuser()]
    try:
        rc, output, error = invoke_command(cmd)
        if rc == 0:
            passwd_reqd = False
            regex = re.compile("must be changed")
            for line in output:
                if line and regex.search(line):
                    passwd_reqd = True
        else:
            lg.error(
                "Error checking user password expiration, rc: %s, out: %s, "
                "err: %s", rc, output, error, error_code=ec.OPERATION_FAILED)
    except Exception:
        lg.exception("Error checking user password expiration",
                     error_code=ec.INTERNAL_ERROR)
        passwd_reqd = True
    return passwd_reqd


def start_nsx_services():
    _SUDO = "/usr/bin/sudo"
    _CMD_START_NSX_SERVICES = (
        "/opt/vmware/nsx-cli/bin/scripts/start_nsx_services")
    cmd = [_SUDO, _CMD_START_NSX_SERVICES]
    try:
        po = subprocess.Popen(cmd)
        po.wait()
        # The backend script will log "Request to enter runlevel 3" on
        # success.  If we don't see that then there is a problem.
    except Exception:
        lg.error("Unable start NSX services",
                 error_code=ec.INTERNAL_ERROR)


def inactivity_timeout_handler():
    sys.stdout.write("\n")
    if lg:
        lg.info(
            "NSX CLI stopped due to inactivity timeout for user: %s" %
            getpass.getuser())
    render_error("MSG_CLI_EXITING_DUE_TO_INACTIVITY")
    sys.exit(CMD_SESSION_TIMEOUT)


def main():
    config = _get_config()
    raw_output = config["raw_output"]
    # Configure logging, if log_file_path not specified then use syslog.
    # The log file is only specified in a development environment, or when
    # nsx-cli is running on ESXi and KVM hosts where we can't assume syslog
    # is running.
    global lg
    log_file_path = config.get("log_file_path")
    log_level = config.get("log_level") or nsx_logging.INFO
    _config_logger(log_file_path, "cli", log_level)
    lg = nsx_logging.getLogger("cli")
    lg.info(
        "NSX CLI started (%s) for user: %s" %
        (get_node_type_string(), getpass.getuser()))

    # Determine if CLI commands are entered via a tty or via stdin redirection
    is_tty = False
    if sys.stdin is not None:
        is_tty = os.isatty(sys.stdin.fileno())

    if is_appliance():
        # Restrict the usage of nsxcli for disallowed users (root,) on appliance nodes
        current_user = getpass.getuser()
        if is_tty and current_user in DISALLOWED_TTY_USERS:
            _error = ("User '%s' is not allowed to invoke CLI in interactive "
                      "mode. Please 'su' as one of the supported users (e.g. %s"
                      ") to use the CLI." % (current_user, _get_supported_users())
                     )
            lg.error(_error)
            sys.stdout.write(_error + "\n")
            sys.exit(CMD_UNSUPPORTED_MODE)

    cli_prompt = CliPrompt()
    user_perms = _get_user_permissions()
    central_cli_enabled = should_enable_central_cli()

    # _NODE_TYPE can have muliple space-delimited values for the all-in-one
    # appliance, we create a list of applicable top level files here to pass
    # to the descriptor factory to load applicable descriptors per node type.
    top_level_ymls = []
    try:
        for node_type in _NODE_TYPE.split():
            top_level_yml = _NODE_TYPE_YML_MAP.get(node_type)
            if not top_level_yml:
                lg.error(
                         "Unknown type '%s' in _NODE_TYPE '%s'",
                         node_type, _NODE_TYPE,
                         error_code=ec.OPERATION_NOT_SUPPORTED)
                continue
            yml_path = os.path.join(_TOP_LEVEL_YML_DIR, top_level_yml)
            if not os.path.isfile(yml_path):
                lg.error("Top level yml file '%s' not found", yml_path,
                         error_code=ec.OPERATION_FAILED)
                continue
            top_level_ymls.append(yml_path)
        if not top_level_ymls:
            lg.error("Yml files not found for node type: '%s'", _NODE_TYPE,
                     error_code=ec.OPERATION_FAILED)
            sys.exit(CMD_UNEXPECTED_ERROR)
        descriptor_factory = DescriptorFactory(
            top_level_ymls,
            settings_file=None,
            user_permissions=user_perms,
            enable_central_cli=central_cli_enabled)
    except SystemExit as ex:
        lg.exception("Unexpected exit:", error_code=ec.INTERNAL_ERROR)
        lg.info("NSX CLI stopped for user: %s" % getpass.getuser())
        sys.exit(ex.code)
    service = CommandService(descriptor_factory, cli_prompt)

    context = MODE_ENABLE
    config['env'] = {}

    # Used while interpreting Central CLI commands
    if "cmd" in config and config["cmd"] == []:
        status = CMD_EXECUTED
        try:
            try:
                stdin_input = sys.stdin.read()
                input_dict = json.loads(stdin_input)
            except Exception:
                lg.exception("Error loading stdin input",
                             error_code=ec.INTERNAL_ERROR)
                sys.exit(CMD_UNEXPECTED_EXCEPTION)
            if input_dict.get("input_type") == "invoke_cmd":
                if ("command" not in input_dict):
                    lg.error("Invalid stdin input, missing command",
                             error_code=ec.INVALID_ARGUMENT)
                    sys.exit(CMD_UNEXPECTED_ERROR)
                cmd_tab_items = input_dict.get("command").split('\t')

                if len(cmd_tab_items) == 2:
                    env_str = cmd_tab_items[0]
                    config['env'] = dict(
                        j.split('=') for j in [i for i in env_str.split()])
                    config['cmd'] = cmd_tab_items[1].split()
                else:
                    config['cmd'] = cmd_tab_items[0].split()
            else:
                status, output = service.process_stdin_arg_request(input_dict)
                if status != CMD_EXECUTED:
                    sys.exit(status)
                elif output:
                    sys.stdout.write(output)
                    sys.stdout.write("\n")
                sys.exit(status)
        except SystemExit as status:
            lg.info("NSX CLI stopped for user: %s" % getpass.getuser())
            sys.exit(status)
    if "cmd" in config and config["cmd"]:
        cmd_string = " ".join(config["cmd"])
        # Support multiple commands delimited by semicolons, e.g.:
        # ./nsxcli --cmd "get version ; set hostname fubar"
        status = CMD_EXECUTED
        cmds = cmd_string.split(";")

        # CENTRAL_CLI_STR existence in cmd['env'] signals that context/navigate
        # command should display their output
        if CENTRAL_CLI_STR in config['env']:
            service._output_navigate_cmd = True
        else:
            service._output_navigate_cmd = False

        cmd = [cmd.strip() for cmd in cmds]

        # Add Central CLI descriptors, if any command is Central CLI command
        if (central_cli_enabled and
                any([is_central_cli_command(cmd.split()) for cmd in cmds])):
            descriptor_factory.process_central_cli_descriptors(service)
            descriptor_factory.node_cache  # Force load registered nodes

        for cmd in cmds:
            result, context = service.invoke_command(cmd, raw_output, context)
            if result == CMD_EXECUTED_REQUESTED_EXIT:
                break
            if result in [CMD_NOT_EXECUTED, CMD_EXECUTED_WITH_ERROR_RESULT]:
                status = result
                break
        lg.info("NSX CLI stopped for user: %s" % getpass.getuser())
        sys.exit(status)

    if is_container():
        sys.stdout.write(
            "NSX CLI (%s). Press ? for command list or enter: help\n" %
            get_node_type_string())
    elif _NODE_TYPE not in [
            NODE_TYPE_HOST_SLESKVM,
            NODE_TYPE_HOST_CENTOSKVM,
            NODE_TYPE_HOST_ESX, NODE_TYPE_HOST_RHELKVM,
            NODE_TYPE_HOST_UBUNTUKVM, NODE_TYPE_HOST_WIN,
            NODE_TYPE_HOST_PUBLIC_CLOUD,
            NODE_TYPE_BAREMETAL_UBUNTU_SERVER,
            NODE_TYPE_BAREMETAL_RHEL_SERVER,
            NODE_TYPE_BAREMETAL_CENTOS_SERVER,
            NODE_TYPE_BAREMETAL_SLES_SERVER,
            NODE_TYPE_BAREMETAL_OEL_SERVER,
            NODE_TYPE_BAREMETAL_WINDOWS_SERVER,
            NODE_TYPE_BAREMETAL_RHEL_CONTAINER,
            NODE_TYPE_BAREMETAL_CENTOS_CONTAINER]:

        if is_tty and not admin_passwd_required():
            # The backend script in start_nsx_services creates this file.
            if not os.path.isfile('/config/.bootstrap_admin_done'):
                start_nsx_services()

        # pam_motd.so is typically used to display the motd on linux
        # distros. However, it doesn't look like the unified Ubuntu
        # 14.04 (trusty) appliance is configured to use it. So we
        # print the contents of /etc/motd.tail manually for now.
        _print_motd_tail(lg)
        _print_upgrade_status(lg)
        _print_bootup_error(lg)
        _print_secondary_disk_error(lg)
        _print_sw_integrity_check_error(lg)
        # Local REST API to get NSX version is not supported on end hosts
        version = _get_nsx_version(lg) or ""
        node_info = (
            "%s %s" % (get_node_type_string(), version)
            if version else get_node_type_string())
        sys.stdout.write(
            "NSX CLI (%s). Press ? for command list or enter: help\n" %
            node_info.title())
    # Configure inactivity timeout handler for controllers/edges/managers
    timeout_handler = inactivity_timeout_handler if is_appliance() else None
    # input_utils.InputHelper needs to be imported after processing the --cmd
    # path which is used by QE automation testcases. For some reason, ansible
    # invocations using --cmd encounters errors if input_utils.InputHelper is
    # imported. Therefore, we delay importing it until now.
    from cli.utils.input_utils import InputHelper
    input_helper = InputHelper(cli_prompt, service, timeout_handler)

    # Start API caching thread if CentralCLI is enabled (on manager node)
    if central_cli_enabled:
        # Used to signal the CLI exit, to gracefully stop the thread
        stop_event = threading.Event()
        central_cli_cache_thread = CentralCLICacheThread(
            descriptor_factory=descriptor_factory, stop_event=stop_event,
            service=service)
        central_cli_cache_thread.start()

    try:
        while True:
            # Ensure parent proccess has not died which will cause a loop of
            # EOFErrors or IOErrors.
            if not is_windows() and os.getppid() == 1:
                try:
                    lg.error("Unexpected parent process pid (1), aborting CLI",
                             error_code=ec.INTERNAL_ERROR)
                except:
                    pass
                sys.exit(CMD_UNEXPECTED_ERROR)
            try:
                try:
                    user_input = input_helper.get_input(context).strip()
                    # If CLI commands entered via stdin redirection, output
                    # the user_input. (This is a useability enhancement for QE.
                    # If we don't do this, stdout only shows the command output.)
                    if not is_tty:
                        sys.stdout.write("%s\n" % user_input)
                except IOError as ex:
                    if ex.errno == errno.EPIPE:
                        continue
                    raise
                if not user_input or user_input[0] in COMMENT_CHARS:
                    continue
                result, context = service.invoke_command(
                    user_input, raw_output, context)
                if result == CMD_EXECUTED_REQUESTED_EXIT:
                    break
            except KeyboardInterrupt:
                sys.stdout.write("\n")
            except EOFError:
                # Client has closed the connection, if stdout still available,
                # print newline, otherwise, ignore errors and exit
                try:
                    sys.stdout.write("\n")
                except Exception:
                    pass
                break
            except Exception as ex:
                if (isinstance(ex, IOError) and hasattr(ex, "errno") and
                        ex.errno == errno.EIO):
                    lg.warning("NSX CLI exited during IO operation")
                    break
                else:
                    lg.exception("Unexpected exception:\n",
                                 error_code=ec.INTERNAL_ERROR)
                    render_error("MSG_CLI_INTERNAL_ERROR")
    finally:
        if central_cli_enabled:
            # Signal the API cache thread to stop
            stop_event.set()

    lg.info("NSX CLI stopped for user: %s" % getpass.getuser())


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(CLI_INTERRUPTED_WHILE_LOAD)
    except Exception:
        try:
            lg.exception("Unexpected error:", error_code=ec.INTERNAL_ERROR)
        except Exception:
            # An error may have occurred before logging is configured
            traceback.print_exc()
        sys.exit(CMD_UNEXPECTED_EXCEPTION)
    finally:
        if is_appliance():
            stop_all_packet_capture_to_file()
