# **********************************************************************
# Copyright 2020-2021 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import importlib
import os
import tempfile
import re
import subprocess
import sys
sys.path.append("/opt/vmware/nsx-common/python")
sys.path.append('/opt/vmware/snmp/lib')
import nsx_logging  # noqa: E402
import cfg  # noqa: E402
import vmw_snmp  # noqa: E402

from nsx_utils.service_utils import service_action  # noqa: E402
from nsx_utils.service_utils import service_status  # noqa: E402
from nsx_utils.service_utils import systemctl_action  # noqa: E402
from nsx_utils.node_utils import is_appliance  # noqa: E402

_PASSWORD_COMPLEXITY_CHECK = False

if is_appliance():
    from nsx_utils.pam_utils import _set_password
    _PASSWORD_COMPLEXITY_CHECK = True

httplib_module = "http.client" if sys.version_info >= (3,) else "httplib"
httplib = importlib.import_module(httplib_module)

SCRIPT_NAME = "vmware-snmpd"
PROCESS_NAME = "snmpd"
SERVICE_NAME = "snmp"

_CONF_NAME = "/etc/vmware/snmp/snmp.xml"

_DEFAULT_SECURTY_LEVEL = "AUTH_PRIV"
_DEFAULT_V2_ACCESS = "read_only"
_DEFAULT_V3_ACCESS = "READ_ONLY"

_RW_COMMUNITY_RE = re.compile("^[\s]*rocommunity[\s]+([\S]{1,64})\s.*$")
_RW_COMMUNITY_COMMENT_RE = re.compile("^[\s]*#[\s]*rocommunity[\s].*$")
_SNMPD_LINK_RE = re.compile("S[0-9]+%s" % SCRIPT_NAME)

_V1V2C_COMMUNITIES = "V1/V2c Communities"
_V2C_NOTIFICATION_TARGETS = "V2c Notification targets"
_V3_NOTIFICATION_TARGETS = "V3 Notification targets"
_V3_USERS = 'V3 Users'


lg = nsx_logging.getLogger(__name__)


# Sample output of vicfg-snmp config
#
# root@manager1:~# vicfg-snmp -s
# Current SNMP agent setting
# Enabled                 : true
# UDP port                : 161
# V1/V2c Communities      : public1 public2
# V2c Notification targets : 10.10.10.10@1162 public3 20.20.20.20@1162 public1
# Notification filter oids:
# V3 Notification targets : 10.10.10.10@1162 user1 authPriv 1
#                           10.10.10.11@1162 user2 authPriv 1
# Contact                 :
# Location                :
# Engine ID               : 80001adc803a89862b0a351e5e00000000
# Auth Protocol           : usmHMACSHAAuthProtocol
# Priv Protocol           : usmAESCfb128PrivProtocol
# V3 Users                : user1 fa8ba6e39a12ebacd35c70acefd8d4d1e449f93d
#                           fa8ba6e39a12ebacd35c70acefd8d4d1e449f93d authPriv
#                           user2 fa8ba6e39a12ebacd35c70acefd8d4d1e449f93d
#                           fa8ba6e39a12ebacd35c70acefd8d4d1e449f93d authPriv
# Log level               : warning
# Process ID              : 29889
# INFO: listing complete.

def snmp_start_on_boot():
    """Return True if snmpd service start_on_boot is set; otherwise, False."""
    links = os.listdir("/etc/rc3.d")
    for link in links:
        if _SNMPD_LINK_RE.match(link) is not None:
            return True
    return False


def get_is_v2_configured(parsed_snmp_configs):
    return True if parsed_snmp_configs.get(_V1V2C_COMMUNITIES) else False


def get_v2_communities(parsed_snmp_configs):
    if not parsed_snmp_configs.get(_V1V2C_COMMUNITIES):
        return []
    communities_json_val = []
    communities = parsed_snmp_configs[_V1V2C_COMMUNITIES].split(" ")
    for community in communities:
        communities_json_val.append({"community_string": community,
                                     "access": _DEFAULT_V2_ACCESS})
    return communities_json_val


def get_v2_targets_without_sensitive_data(parsed_snmp_configs):
    if not parsed_snmp_configs.get(_V2C_NOTIFICATION_TARGETS):
        return []
    v2_targets_json_val = []
    v2_server_ports =\
        parsed_snmp_configs[_V2C_NOTIFICATION_TARGETS].split(" ")[::2]
    for server_port in v2_server_ports:
        v2_targets_json_val.append({"server": server_port.split('@')[0],
                                    "port": int(server_port.split('@')[1])})
    return v2_targets_json_val


def get_v2_targets_with_sensitive_data(parsed_snmp_configs):
    if not parsed_snmp_configs.get(_V2C_NOTIFICATION_TARGETS):
        return []
    v2_targets_json_val = []
    v2_server_ports =\
        parsed_snmp_configs[_V2C_NOTIFICATION_TARGETS].split(" ")[::2]
    v2_communities =\
        parsed_snmp_configs[_V2C_NOTIFICATION_TARGETS].split(" ")[1::2]
    for server_port, community in zip(v2_server_ports, v2_communities):
        v2_targets_json_val.append({"server": server_port.split('@')[0],
                                    "port": int(server_port.split('@')[1]),
                                    "community_string": community})
    return v2_targets_json_val


def get_is_v3_configured(parsed_snmp_configs):
    if _V3_USERS not in parsed_snmp_configs:
        return False
    v3_user_id = get_v3_user_id(parsed_snmp_configs[_V3_USERS])
    is_v3_configured = (True if v3_user_id != "" else False)
    return is_v3_configured


def should_snmpd_be_running(req_obj):
    v2_communities = (True if req_obj.get("communities") else False)
    v2_targets = (True if req_obj.get("v2_targets") else False)
    v3_users = (True if req_obj.get("v3_users") else False)
    return v2_communities or v3_users or v2_targets


def get_v3_auth_protocol(parsed_snmp_configs):
    return vmw_snmp._convert_auth_proto_to_pint(
        parsed_snmp_configs['Auth Protocol'])


def get_v3_priv_protocol(parsed_snmp_configs):
    return vmw_snmp._convert_priv_proto_to_pint(
        parsed_snmp_configs['Priv Protocol'])


def get_v3_users(parsed_snmp_configs):
    if _V3_USERS not in parsed_snmp_configs:
        return []
    v3_user_id_list = parsed_snmp_configs[_V3_USERS].split(" ")[::4]
    v3_users_json_val = []
    if not v3_user_id_list[0]:
        return v3_users_json_val
    for user_id in sorted(v3_user_id_list):
        v3_users_json_val.append({"user_id": user_id,
                                  "security_level": _DEFAULT_SECURTY_LEVEL,
                                  "access": _DEFAULT_V3_ACCESS})
    return v3_users_json_val


def get_v3_targets(parsed_snmp_configs):
    if _V3_NOTIFICATION_TARGETS not in parsed_snmp_configs:
        return []
    v3_server_ports = parsed_snmp_configs[
        _V3_NOTIFICATION_TARGETS].split(" ")[::4]
    v3_users = parsed_snmp_configs[
        _V3_NOTIFICATION_TARGETS].split(" ")[1::4]

    v3_targets_json_val = []
    if not v3_server_ports[0]:
        return v3_targets_json_val

    for server_port, user_id in zip(v3_server_ports, v3_users):
        v3_targets_json_val.append({"server": server_port.split('@')[0],
                                    "port": int(server_port.split('@')[1]),
                                    "user_id": user_id,
                                    "security_level": _DEFAULT_SECURTY_LEVEL})
    return v3_targets_json_val


def get_snmp_configs():
    po = subprocess.Popen([sys.executable, "/opt/vmware/bin/vicfg-snmp", "-s"],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = [p.decode('utf-8') for p in po.communicate()]
    if po.returncode != 0:
        lg.warning(
            "Error when excuting command %s /opt/vmware/bin/vicfg-snmp -s, "
            "return code: %s, output: %s, error message: %s",
            sys.executable, po.returncode, out, err)
    return out


def get_v3_user_id(parsed_v3_users):
    return parsed_v3_users.split(' ')[0]


def parse_snmp_configs(snmp_configs):
    snmp_configs_by_line = snmp_configs.split('\n')
    parsed_snmp_configs = {}

    for snmp_config_by_line in snmp_configs_by_line:
        snmp_configs_splited = snmp_config_by_line.split(':')
        if len(snmp_configs_splited) is 2:
            parsed_snmp_configs.update({snmp_configs_splited[0].strip():
                                        snmp_configs_splited[1].strip()})
    return parsed_snmp_configs


def set_snmp_service_properties(start_on_boot):
    """Set or unset snmp service start_on_boot.
    Args:
        start_on_boot: True if snmp service start_on_boot configured;
            otherwise, False.
    """
    if start_on_boot:
        cmd = ["/bin/systemctl", "enable", "vmware-snmpd"]
    else:
        cmd = ["/bin/systemctl", "disable", "vmware-snmpd"]
    po = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = [p.decode('utf-8') for p in po.communicate()]
    if po.returncode != 0:
        lg.warning(
            "Error %s %s init script links, rc: %s, out: %s, err: %s",
            "adding" if start_on_boot else "removing", SCRIPT_NAME,
            po.returncode, out, err)
        return False
    return True


def set_snmp_communities(community_string):
    """Set SNMP community string, return True on success, otherwise, False."""
    if community_string is None or len(community_string) == 0:
        lg.info("Empty the community string")
        cfg.set_communities('')
        return True
    # update /etc/vmware/snmp/snmp.xml for community string
    lg.info("start update /etc/vmware/snmp/snmp.xml for community string")
    cfg.set_communities(community_string)
    return True


def set_v3_engine_id(v3_engine_id):
    """Set SNMP engine id, return True on success, otherwise, False."""

    # update /etc/vmware/snmp/snmp.xml for v3_engine_id
    lg.info("start update /etc/vmware/snmp/snmp.xml for v3_engine_id")
    cfg.set_engineid(v3_engine_id)
    # update /etc/vmware/snmp/snmp.xml to remove the v3user if configured.
    lg.info("start clear /etc/vmware/snmp/snmp.xml to remove the v3user if "
            "configured")
    clear_all_v3users()
    # update /etc/vmware/snmp/snmp.xml to remove the v3target if configured.
    lg.info("start clear /etc/vmware/snmp/snmp.xml to remove the v3target if "
            "configured")
    clear_all_v3targets()
    return True


def set_v3_auth_protocol(v3_auth_protocol):
    """Set SNMP auth protocol, return True on success, otherwise, False."""

    # update /etc/vmware/snmp/snmp.xml for v3_auth_protocol
    lg.info("start update /etc/vmware/snmp/snmp.xml for v3_auth_protocol")
    cfg.set_authentication(v3_auth_protocol)
    return True


def set_v3_priv_protocol(v3_priv_protocol):
    """Set SNMP priv protocol, return True on success, otherwise, False."""

    # update /etc/vmware/snmp/snmp.xml for v3_priv_protocol
    lg.info("start update /etc/vmware/snmp/snmp.xml for v3_priv_protocol")
    cfg.set_privacy(v3_priv_protocol)
    return True


def _get_v3_user_internal_fmt(v3_user, v3_auth_protocol,
                              v3_priv_protocol):
    """Set SNMP user, return True on success, otherwise, False."""
    user_id = v3_user['user_id']
    auth_password = v3_user['auth_password']
    priv_password = v3_user['priv_password']

    # make hash according to eid, auth_protocol, private_protocol
    # write auth_password and priv_password to file
    if sys.version_info >= (3,):
        auth_password = auth_password.encode()
        priv_password = priv_password.encode()

    auth_password_path = None
    priv_password_path = None

    try:
        auth_password_fd, auth_password_path =\
            tempfile.mkstemp(prefix="auth_password")
        os.write(auth_password_fd, auth_password)
        os.close(auth_password_fd)

        priv_password_fd, priv_password_path =\
            tempfile.mkstemp(prefix="priv_password")
        os.write(priv_password_fd, priv_password)
        os.close(priv_password_fd)

        po = subprocess.Popen(["/opt/vmware/bin/snmpd", "-a", "SHA1",
                            "-s", auth_password_path,
                            "-p", "AES128", "-s", priv_password_path],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = [p.decode('utf-8') for p in po.communicate()]
        if po.returncode != 0:
            lg.warning(
                "Error when excuting command /opt/vmware/bin/snmpd -a %s -s "
                "/tmp/auth_password -p %s -s /tmp/priv_password, "
                "return code: %s, output: %s, error message: %s",
                v3_auth_protocol, v3_priv_protocol, po.returncode, out, err)
            return False, None

        # get the auth_hash and priv_hash from the output of the cmd
        auth_hash = re.findall(".*<auth>(.*)</auth>.*", out)[0]
        priv_hash = re.findall(".*<priv>(.*)</priv>.*", out)[0]
        return True, [user_id, auth_hash, priv_hash, "priv"]
    finally:
        try:
            if auth_password_path:
                os.remove(auth_password_path)
        except Exception as e:
            lg.warning("Error when removing temporary auth password file."
                "ERROR: %s", e)
        try:
            if priv_password_path:
                os.remove(priv_password_path)
        except Exception as e:
            lg.warning("Error when removing temporary priv password file."
                "ERROR: %s", e)


def clear_all_v3users():
    f = open(_CONF_NAME, 'r')
    snmp_conf = f.read()
    v3_old_users = re.findall(".*<v3users>(.*)</v3users>.*", snmp_conf)[0]
    new_snmp_conf = snmp_conf.replace(v3_old_users, "")

    # override the snmp.xml file
    with open(_CONF_NAME, 'w') as f:
        f.write(new_snmp_conf)
    f.close()


def clear_all_targets():
    f = open(_CONF_NAME, 'r')
    snmp_conf = f.read()
    v3_old_users = re.findall(".*<targets>(.*)</targets>.*", snmp_conf)[0]
    new_snmp_conf = snmp_conf.replace(v3_old_users, "")

    # override the snmp.xml file
    with open(_CONF_NAME, 'w') as f:
        f.write(new_snmp_conf)


def clear_all_v3targets():
    f = open(_CONF_NAME, 'r')
    snmp_conf = f.read()
    v3_old_users = re.findall(".*<v3targets>(.*)</v3targets>.*", snmp_conf)[0]
    new_snmp_conf = snmp_conf.replace(v3_old_users, "")

    # override the snmp.xml file
    with open(_CONF_NAME, 'w') as f:
        f.write(new_snmp_conf)
    f.close()


def validate_start_on_boot(req_obj, parsed_snmp_configs):
    start_on_boot = req_obj.get("start_on_boot")
    if start_on_boot and not should_snmpd_be_running(req_obj):
        return (False, httplib.CONFLICT,
                "NODE_ERROR_SNMP_SERVICE_START_REQUIRES_COMMUNITY_OR_V3_USER")
    return True, None, None


def update_start_on_boot(req_obj, parsed_snmp_configs):
    start_on_boot = req_obj.get("start_on_boot",
        should_snmpd_be_running(req_obj))
    # Configure start_on_boot
    if not set_snmp_service_properties(start_on_boot):
        return (False, httplib.INTERNAL_SERVER_ERROR,
                "NODE_ERROR_SNMP_SERVICE_UPDATE_CONF_FAILED")
    return True, None, None


def validate_communities(req_obj, parsed_snmp_configs):
    # Configure communities if specified in request
    communities = req_obj.get("communities", [])

    for community in communities:
        if not community:
            continue
        community_string = community.get("community_string")
        if not community_string:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_COMMUNITY_STRING_INVALID")
        if community_string is not None:
            is_communities_valid = re.match('^.{1,64}$', community_string)
            if is_communities_valid is None:
                return (False, httplib.BAD_REQUEST,
                        "NODE_ERROR_SNMP_COMMUNITY_STRING_INVALID")

    return True, None, None


def update_communities(req_obj, parsed_snmp_configs):
    # Configure communities if specified in request
    communities_string = ""
    communities = req_obj.get("communities", [])

    for community in communities:
        if not community:
            continue
        community_string = community.get("community_string")
        if community_string is not None:
            if communities_string:
                communities_string += "," + community_string
            else:
                communities_string += community_string

    if not set_snmp_communities(communities_string):
        return (False, httplib.INTERNAL_SERVER_ERROR,
                "NODE_ERROR_SNMP_SERVICE_UPDATE_CONF_FAILED")

    lg.info("Updated communities in %s successfully", _CONF_NAME)
    return True, None, None


def get_existing_matched_v2_target(cur_v2_target,exist_v2_targets ):
    lg.info("Check if v2_target: %s@%s is a existing v2_target"
        , cur_v2_target["server"], str(cur_v2_target["port"]))
    for exist_v2_target in exist_v2_targets:
        if exist_v2_target["server"] == cur_v2_target["server"] \
                and exist_v2_target["port"] == cur_v2_target["port"]:
            return exist_v2_target
    return None


def validate_v2_targets(req_obj, parsed_snmp_configs):
    v2_targets = req_obj.get("v2_targets", [])
    exist_v2_targets = get_v2_targets_with_sensitive_data(parsed_snmp_configs)
    for v2_target in v2_targets:
        # when community_string is not in the request body, check if it is an
        # existing one, if not, throw an error; if yes no need to do any
        # further validation
        if 'community_string' not in v2_target:
            exist_v2_target = get_existing_matched_v2_target(v2_target,
                exist_v2_targets)
            if exist_v2_target is None:
                return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_COMMUNITY_STRING_IN_V2_TARGETS_MISSING")
            continue
        else:
            community_string = v2_target["community_string"]
        is_communities_valid = re.match('^.{1,64}$', community_string)
        if is_communities_valid is None:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_COMMUNITY_STRING_IN_V2_TARGETS_INVALID")
    return True, None, None


def update_v2_targets(req_obj, parsed_snmp_configs):
    targets_str = ""
    debug_targets_str = ""
    v2_targets = req_obj.get("v2_targets", [])
    exist_v2_targets = get_v2_targets_with_sensitive_data(parsed_snmp_configs)
    for v2_target in v2_targets:
        #when community_string is not in the request body,
        #try to get the corresponding one from the existing v2_targets
        if 'community_string' not in v2_target:
            exist_v2_target = get_existing_matched_v2_target(v2_target,
                exist_v2_targets)
            community_string = exist_v2_target["community_string"]
        else:
            community_string = v2_target["community_string"]

        target_str = v2_target["server"] + "@" + str(v2_target["port"]) +\
                     "/" + community_string
        debug_target_str = v2_target["server"] + "@" + str(v2_target["port"])

        if targets_str:
            targets_str += "," + target_str
            debug_targets_str += "," + debug_target_str
        else:
            targets_str = target_str
            debug_targets_str = debug_target_str
    lg.info("Start updating %s for v2_targets %s", _CONF_NAME,
        debug_targets_str)
    if targets_str:
        cfg.set_targets(targets_str)
        cfg.set_v2c('true')
        lg.info('set snmp v2c flag to true')
    else:
        clear_all_targets()

    return True, None, None


def validate_v3_protocols(req_obj, parsed_snmp_configs):
    v3_auth_protocol = req_obj.get("v3_auth_protocol", "")
    v3_priv_protocol = req_obj.get("v3_priv_protocol", "")
    is_auth_protocol_valid = re.match('SHA1', v3_auth_protocol)
    if is_auth_protocol_valid is None:
        return (False, httplib.BAD_REQUEST,
                "NODE_ERROR_SNMP_V3_AUTH_PROTOCOL_INVALID")
    is_priv_protocol_valid = re.match('AES128', v3_priv_protocol)
    if is_priv_protocol_valid is None:
        return (False, httplib.BAD_REQUEST,
                "NODE_ERROR_SNMP_V3_PRIV_PROTOCOL_INVALID")
    return True, None, None


def update_v3_protocols(req_obj, parsed_snmp_configs):
    old_v3_auth_protocol = get_v3_auth_protocol(parsed_snmp_configs)
    old_v3_priv_protocol = get_v3_priv_protocol(parsed_snmp_configs)
    v3_auth_protocol = req_obj["v3_auth_protocol"]
    v3_priv_protocol = req_obj["v3_priv_protocol"]

    # Only need to update, if the new value if different with the old one
    if old_v3_auth_protocol != v3_auth_protocol:
        set_v3_auth_protocol(v3_auth_protocol)

    if old_v3_priv_protocol != v3_priv_protocol:
        set_v3_priv_protocol(v3_priv_protocol)
    return True, None, None


def validate_v3_users(req_obj, parsed_snmp_configs):
    old_v3_user_id_list = []
    if parsed_snmp_configs.get(_V3_USERS):
        old_v3_user_id_list = parsed_snmp_configs[_V3_USERS].split(" ")[::4]

    v3_users = req_obj.get("v3_users", [])
    v3_auth_protocol = req_obj["v3_auth_protocol"]
    v3_priv_protocol = req_obj["v3_priv_protocol"]

    for v3_user in v3_users:
        is_userid_valid = re.match('^.{1,32}$', v3_user['user_id'])
        if is_userid_valid is None:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_V3_USERID_INVALID")
        # We need to override with the new user info
        if v3_user['user_id'] not in old_v3_user_id_list:
            if 'auth_password' in v3_user and 'priv_password' in v3_user \
                    and v3_user['auth_password'] != '' and \
                    v3_user['priv_password'] != '':
                if _PASSWORD_COMPLEXITY_CHECK:
                    # Do the PAM password complexity check
                    res, msg = _set_password("passwduser",
                                             v3_user['auth_password'])
                    if res == 1:
                        lg.info("_set_password returned error for the "
                            "auth_password of v3-user: %s", v3_user['user_id'])
                        return (False, httplib.BAD_REQUEST, "NODE_ERROR_"
                                "SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")
                    res, msg = _set_password("passwduser",
                                             v3_user['priv_password'])
                    if res == 1:
                        lg.info("_set_password returned error for the "
                            "priv_password of v3-user: %s", v3_user['user_id'])
                        return (False, httplib.BAD_REQUEST, "NODE_ERROR_"
                                "SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")
            else:
                return (False, httplib.BAD_REQUEST,
                        "NODE_ERROR_SNMP_SERVICE_PASSWORD_NUMBER_ERROR")
        else:
            # Upgrade the old user (we do not save password, so update it
            # anyway)
            if 'auth_password' in v3_user and 'priv_password' in v3_user \
                   and v3_user['auth_password'] != '' and \
                   v3_user['priv_password'] != '':
                if _PASSWORD_COMPLEXITY_CHECK:
                    # Do the PAM password complexity check
                    res, msg = _set_password("passwduser",
                                             v3_user['auth_password'])
                    if res == 1:
                        lg.info("_set_password returned error for the "
                            "auth_password of v3-user: %s", v3_user['user_id'])
                        return (False, httplib.BAD_REQUEST, "NODE_ERROR_"
                                "SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")
                    res, msg = _set_password("passwduser",
                                             v3_user['priv_password'])
                    if res == 1:
                        lg.info("_set_password returned error for the "
                            "priv_password of v3-user: %s", v3_user['user_id'])
                        return (False, httplib.BAD_REQUEST, "NODE_ERROR_"
                                "SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")
                    lg.info("Validated both v3 passwords")
            elif 'auth_password' not in v3_user and 'priv_password' \
                    not in v3_user:
                lg.info("No need update for user")
            else:
                return (False, httplib.BAD_REQUEST,
                        "NODE_ERROR_SNMP_SERVICE_PASSWORD_NUMBER_ERROR")

    return True, None, None


class v3_user_std():

    def __init__(self):
        self._users = {}

    def _convert_vicfg_to_internal(self, users_str):
        key = None
        arrvalue = []
        for item in users_str.split(" "):
            if len(arrvalue) == 0:
                key = item
            # vicfg gives "authPriv" in output, but expects "priv" in input
            if len(arrvalue) == 3 and item == "authPriv":
                item = "priv"
            arrvalue.append(item)
            if len(arrvalue) >= 4:
                self._users[key] = arrvalue
                key = None
                arrvalue = []

    def read_users_from_vicfg(self):
        snmp_configs = get_snmp_configs()
        users_str = parse_snmp_configs(snmp_configs)[_V3_USERS]
        self._convert_vicfg_to_internal(users_str)

    def read_users_from_parsed_snmp_configs(self, parsed_snmp_configs):
        users_str = parsed_snmp_configs[_V3_USERS]
        self._convert_vicfg_to_internal(users_str)

    def get_user(self, user_id):
        # Get user in internal format.
        # Returns None if not found.
        if user_id not in self._users:
            return None
        return self._users[user_id]

    def insert_user(self, user):
        # Insert a user where the user is provided in our internal format.
        # 'user' is expected to be a list of length 4 where the first item
        # is the user_id, for  example,
        # ["user3", "b62de90e8235fa147c98d7fe0c95edf884a53fb7",
        #  "b62de90e8235fa147c98d7fe0c95edf884a53fb7", "privauthPriv"]

        if not isinstance(user, list):
            return None
        self._users[user[0]] = user
        return self._users[user[0]]

    def insert_user_api_fmt(self, user, v3_auth_protocol, v3_priv_protocol):
        # Insert a user given in our API format, for example,
        # {'security_level': 'AUTH_PRIV', 'access': 'READ_ONLY',
        #  'auth_password': 'Admin!23Admin',
        #  'priv_password': 'Admin!23Admin', 'user_id': 'user4'}
        # The function validates the two passwords, converts the user into
        # our internal format, and inserts into internal dictionary.
        if 'user_id' not in user:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_V3_USERID_INVALID")
        is_userid_valid = re.match('^.{1,32}$', user['user_id'])
        if is_userid_valid is None:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_V3_USERID_INVALID")
        if 'auth_password' not in user or 'priv_password' not in user:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")
        if _PASSWORD_COMPLEXITY_CHECK:
            res, msg = _set_password("passwduser", user['auth_password'])
            if res == 1:
                lg.info("_set_password returned error for the auth_password "
                        "of v3-user: %s", user['user_id'])
                return (False, httplib.BAD_REQUEST,
                        "NODE_ERROR_SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")

            res, msg = _set_password("passwduser", user['priv_password'])
            if res == 1:
                lg.info("_set_password returned error for the priv_password "
                        "of v3-user: %s", user['user_id'])
                return (False, httplib.BAD_REQUEST,
                        "NODE_ERROR_SNMP_SERVICE_PASSWORD_COMPLEXITY_ERROR")

        res, arruser = _get_v3_user_internal_fmt(user, v3_auth_protocol,
                                                 v3_priv_protocol)
        if not res:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_SERVICE_PASSWORD_HASHGEN_ERROR")
        self.insert_user(arruser)
        return True, None, None

    def to_vicfg_string(self):
        str = ""
        for user_id in self._users.keys():
            if str:
                str += ","
            str += " ".join(self._users[user_id])
        return str

    def write_users(self):
        users_str = self.to_vicfg_string()
        lg.info("start update /etc/vmware/snmp/snmp.xml for v3-users %s",
                users_str)
        # Always returns None
        cfg.set_users(users_str)


def update_v3_users(req_obj, parsed_snmp_configs):
    v3_auth_protocol = req_obj["v3_auth_protocol"]
    v3_priv_protocol = req_obj["v3_priv_protocol"]
    v3_users = req_obj.get("v3_users", [])

    old_v3_users = v3_user_std()
    old_v3_users.read_users_from_parsed_snmp_configs(parsed_snmp_configs)
    new_v3_users = v3_user_std()
    new_user_ids = []

    if len(v3_users) == 0:
        clear_all_v3users()
        clear_all_v3targets()
        return True, None, None

    old_v3_users_id_list = []
    if parsed_snmp_configs.get(_V3_USERS):
        old_v3_user_id_list = parsed_snmp_configs[_V3_USERS].split(" ")[::4]

    for v3_user in v3_users:
        if 'user_id' not in v3_user:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_V3_USERID_INVALID")
        user_id = v3_user['user_id']
        if 'auth_password' in v3_user and 'priv_password' in v3_user \
                and v3_user['auth_password'] != '' \
                and v3_user['priv_password'] != '':
            new_user_ids.append(user_id)
            # If insert fails then validation didn't work.
            new_v3_users.insert_user_api_fmt(v3_user, v3_auth_protocol,
                                             v3_priv_protocol)
        else:
            tmp_user = old_v3_users.get_user(user_id)
            if tmp_user:
                new_v3_users.insert_user(tmp_user)

    # persist to vicfg
    new_v3_users.write_users()

    return True, None, None


def validate_v3_targets(req_obj, parsed_snmp_configs):
    v3_targets = req_obj.get("v3_targets", [])
    for v3_target in v3_targets:
        v3_users = []
        if "v3_users" in req_obj:
            v3_users = req_obj["v3_users"]

        v3_user_ids = [d['user_id'] for d in v3_users]
        user_id = v3_target["user_id"]
        if user_id not in v3_user_ids:
            return (False, httplib.BAD_REQUEST,
                    "NODE_ERROR_SNMP_V3_USERID_UNKNOWN", user_id)

    return True, None, None, None


def update_v3_targets(req_obj, parsed_snmp_configs):
    targets_str = ""
    v3_targets = req_obj.get("v3_targets", [])
    for v3_target in v3_targets:
        target_str = v3_target["server"] + "@" + str(v3_target["port"]) +\
                     "/" + v3_target["user_id"] +\
                     "/" + "priv" + "/" + "trap"
        if targets_str:
            targets_str += "," + target_str
        else:
            targets_str = target_str
    lg.info("start update /etc/vmware/snmp/snmp.xml for v3_targets %s",
            targets_str)
    if targets_str:
        cfg.set_v3targets(targets_str)
    else:
        clear_all_v3targets()

    return True, None, None


def validate_snmp_config(req_obj, parsed_snmp_configs):
    res, http_err, err = validate_start_on_boot(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_start_on_boot failed")
        return res, http_err, err, None
    res, http_err, err = validate_communities(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_communities failed")
        return res, http_err, err, None
    res, http_err, err = validate_v2_targets(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_v2_targets failed")
        return res, http_err, err, None
    res, http_err, err = validate_v3_protocols(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_v3_protocols failed")
        return res, http_err, err, None
    res, http_err, err = validate_v3_users(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_v3_users failed")
        return res, http_err, err, None
    res, http_err, err, user_id =\
        validate_v3_targets(req_obj, parsed_snmp_configs)
    if not res:
        lg.warn("validate_v3_targets failed")
        return res, http_err, err, user_id

    return True, None, None, None


def update_v2_v3_configured(req_obj, parsed_snmp_configs):

    rc, out, err = service_status(SCRIPT_NAME)
    if should_snmpd_be_running(req_obj):
        # snmp needs to be started if not started already
        if rc != 0:
            # snmpd is not running, need to start it
            rc, out, err = service_action(SCRIPT_NAME, "restart")
            if rc != 0:
                lg.warning(
                    "Error starting snmp service, rc: %s, out: %s, err: %s",
                    rc, out, err)
                return (False, httplib.INTERNAL_SERVER_ERROR,
                        "NODE_ERROR_SNMP_SERVICE_START_FAILED")
    else:
        # snmp needs to be stopped if not stopped already
        if rc == 0:
            # snmpd is running, need to stop it
            rc, out, err = service_action(SCRIPT_NAME, "stop")
            if rc != 0:
                lg.warning(
                    "Error stopping snmp service, rc: %s, out: %s, err: %s",
                    rc, out, err)
                return (False, httplib.INTERNAL_SERVER_ERROR,
                        "NODE_ERROR_SNMP_SERVICE_STOP_FAILED")
    return True, None, None


def update_snmp_config(req_obj, parsed_snmp_configs):
    res, http_err, err = update_start_on_boot(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_communities(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_v2_targets(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_v3_protocols(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_v3_users(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_v3_targets(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err
    res, http_err, err = update_v2_v3_configured(req_obj, parsed_snmp_configs)
    if not res:
        return res, http_err, err

    return True, None, None
