# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import json
import os
import re
import sha.config
import socket
import sys
import time
from datetime import datetime
import uuid
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache
from subprocess import CalledProcessError, Popen, PIPE

from nsx_utils.exporter_utils import read_exporter_properties
from nsx_utils.node_utils import get_node_type
from nsx_utils.constants import NODE_TYPE_EDGE
from nsx_utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from nsx_utils.constants import NODE_TYPE_MANAGER
from nsx_utils.constants import NODE_TYPE_POLICY
from nsx_utils.constants import NODE_TYPE_CONTROLLER

from nsx_monitoring.clientlibrary.feature_event_type_mappings \
    import AUDIT_LOG_HEALTH_REMOTE_LOGGING_SERVER_ERROR_EVENT_TYPE_ID
# TODO: should  add it in the nsx_utils
NODE_TYPE_UA = "nsx-ua"
NODE_TYPE_GM = "global-manager"

_NODE_TYPE = get_node_type()

if NODE_TYPE_MANAGER in _NODE_TYPE and NODE_TYPE_POLICY in _NODE_TYPE \
        and NODE_TYPE_CONTROLLER in _NODE_TYPE:
    _NODE_TYPE = NODE_TYPE_UA

_SUPPORTED_NODE_TYPES = [
    NODE_TYPE_EDGE,
    NODE_TYPE_PUBLIC_CLOUD_GATEWAY,
    NODE_TYPE_UA,
    NODE_TYPE_GM
]

_SERVER = "server"
_PORT = "port"
_PROTOCOL = "protocol"
_EXPORTER_NAME = "exporter_name"
_TLS_CERT_PEM = "tls_cert_pem"

# TODO: change hostname_or_ip_address_with_port to
# hostname_or_ip_address_with_port_and_port in yml first.
_HOSTNAME_OR_IP_ADDRESS_WITH_PORT_AND_PROTOCOL = \
    "hostname_or_ip_address_with_port"
_ENTITY_ID = "entity_id"

_IP_RE_TEXT = \
    "^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$"
_IP_PATTERN = re.compile(_IP_RE_TEXT)
_GET_FQDN_RESOLVE_RES_CMD = ["getent", "ahostsv4"]
_GET_CERT_ENDDATE_CMD = ["/usr/bin/openssl", "x509", "-enddate", "-noout"]
_OPENSSL_DATE_FORMAT = "%b %d %H:%M:%S %Y GMT"
_IPTABLES_USER = "syslog"

_RESTART_RSYSLOG = ["service", "rsyslog", "restart"]
_DEFAULT_UDP_PORT = "514"
_DEFAULT_TCP_PORT = "6514"
_DEFAULT_UDP_IPT_RULE_CHECK_CMD_STR = "iptables -C OUTPUT -w 5 -o eth0 -p udp \
        -m udp --dport {} -m owner --uid-owner syslog \
        -j ACCEPT".format(_DEFAULT_UDP_PORT)
_DEFAULT_UDP_IPT_RULE_CHECK_CMD = _DEFAULT_UDP_IPT_RULE_CHECK_CMD_STR.split()
_DEFAULT_TCP_IPT_RULE_CHECK_CMD_STR = "iptables -C OUTPUT -w 5 -o eth0 -p tcp \
        -m tcp --dport {} -m state --state NEW,ESTABLISHED \
        -m owner --uid-owner syslog -j ACCEPT".format(_DEFAULT_TCP_PORT)
_DEFAULT_TCP_IPT_RULE_CHECK_CMD = _DEFAULT_TCP_IPT_RULE_CHECK_CMD_STR.split()


class AuditLogHealthForRemoteErrorStatus(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._logging_server_status_dict = {}

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        self._logger.debug("node_type: {}".format(_NODE_TYPE))
        # don't need to check if _NODE_TYPE is not in the _SUPPORTED_NODE_TYPES
        if _NODE_TYPE not in _SUPPORTED_NODE_TYPES:
            return {}, []
        # get logging_server list
        logging_server_list = read_exporter_properties() or []
        self._logger.debug("logging_server_list: {}".format(
            logging_server_list))
        logging_server_status_dict = {}
        for cur_logging_server in logging_server_list:
            key = uuid.UUID(cur_logging_server[_EXPORTER_NAME])
            context_data_in_dict = {
                _HOSTNAME_OR_IP_ADDRESS_WITH_PORT_AND_PROTOCOL:
                    cur_logging_server[_SERVER] +
                    "@" + str(cur_logging_server[_PORT]) +
                    "@" + cur_logging_server[_PROTOCOL],
                _ENTITY_ID: cur_logging_server[_EXPORTER_NAME]}
            # do the FQDN resolve check if needed
            if _IP_PATTERN.match(cur_logging_server[_SERVER]) is None:
                is_fqdn_unresolvable_res = self.is_fqdn_unresolvable(
                        cur_logging_server[_SERVER])
                if is_fqdn_unresolvable_res is None:
                    continue
                if is_fqdn_unresolvable_res:
                    logging_server_status_dict[key] = \
                        (True, context_data_in_dict)
                    continue
            # do the cert expire check if needed
            if cur_logging_server[_PROTOCOL] == 'TLS':
                is_cert_expire_res = self.is_cert_expire(
                        cur_logging_server[_TLS_CERT_PEM])
                if is_cert_expire_res is None:
                    continue
                if is_cert_expire_res:
                    logging_server_status_dict[key] = \
                        (True, context_data_in_dict)
                    continue
            logging_server_status_dict[key] = \
                        (False, context_data_in_dict)
        # do the iptable check
        logging_server_status_dict = self.iptable_check(logging_server_status_dict)
        # compare the logging_server_status_dict with the old one
        # (_logging_server_status_dict).
        # add the deleted entities into deleted_entity_list and call
        # event_source.update_deleted_entities to do the deletion.
        deleted_entity_list = []
        need_rsyslog_restart = False
        for old_logging_server_id, old_logging_server_status \
                in list(self._logging_server_status_dict.items()):
            if old_logging_server_id \
                    not in list(logging_server_status_dict.keys()):
                deleted_entity_list.append(old_logging_server_id)
        # check if rsyslog_restart is needed
        for cur_logging_server_id, cur_logging_server_status \
                in list(logging_server_status_dict.items()):
            # if the logging_server is newly addded and error status= true,
            # we should set need_rsyslog_restart to true
            if cur_logging_server_id not in list(
                    self._logging_server_status_dict.keys()):
                if cur_logging_server_status[0]:
                    need_rsyslog_restart = True
                    break
            # if the logging_server is not newly added
            # but the error status changed,
            # we should set need_rsyslog_restart to true
            else:
                if cur_logging_server_status[0] != \
                    self._logging_server_status_dict[cur_logging_server_id][0]:
                    need_rsyslog_restart = True
                    break
        # reload rsyslog config
        if need_rsyslog_restart:
            restart_rsyslog_res = self.restart_rsyslog()
            # if fail to do the restart,
            # not update the logging_server_status dict
            if not restart_rsyslog_res:
                return self._logging_server_status_dict, deleted_entity_list
        self._logger.debug("last logging_server_status_dict: {}".format(
            self._logging_server_status_dict))
        # store it
        self._logging_server_status_dict = logging_server_status_dict
        self._logger.debug("logging_server_status_dict: {}".format(
            logging_server_status_dict))
        return logging_server_status_dict, deleted_entity_list

    def is_fqdn_unresolvable(self, fqdn_address):
        try:
            fqdn_resolve_res = run_command(_GET_FQDN_RESOLVE_RES_CMD +
                [fqdn_address], shell=False) or ''
        except CalledProcessError as e:
            self._logger.debug("{} can't be resolved. Error: {}".format(
                fqdn_address, e.stderr))
            return True
        except Exception as e:
            self._logger.warning(
                "Error executing get fqdn resolve command: {}. "
                    "Error: {}".format(" ".join(_GET_FQDN_RESOLVE_RES_CMD +
                    [fqdn_address]), e.stderr))
            return None
        return False

    def restart_rsyslog(self):
        self._logger.debug("restart rsyslog")
        try:
            fqdn_resolve_res = run_command(_RESTART_RSYSLOG, shell=False) or ''
        except Exception as e:
            self._logger.debug("rsyslog fails to restart. Error: {}".format(
                e.stderr))
            return False
        return True

    def iptable_check(self, logging_server_status_dict):
        for cur_id, cur_status in list(logging_server_status_dict.items()):
            # only need to check if current error_status = false
            # error status: cur_status[0]
            # context data: cur_status[1]
            if not cur_status[0]:
                cur_hostname_or_ip_with_port_and_protocol = \
                    cur_status[1]\
                        [_HOSTNAME_OR_IP_ADDRESS_WITH_PORT_AND_PROTOCOL]
                hostNameOrIp, port, protocol = \
                    self.getHostNameOrIpAndPortAndProtocol(
                        cur_hostname_or_ip_with_port_and_protocol)
                ip_addr = socket.gethostbyname(hostNameOrIp)
                # the iptable rule is okay, if any of the following rules exist
                ipt_rule_exists_res = self.ipt_rule_exists(port,
                        protocol, server=ip_addr, user=_IPTABLES_USER,
                        interface="*") \
                        or self.default_ipt_rule_exists(port)
                if ipt_rule_exists_res is None:
                    continue
                if not ipt_rule_exists_res:
                    # if status changed, reload the rsyslog config
                    logging_server_status_dict[cur_id] = \
                        (True, cur_status[1])
                    self._logger.debug("{}'s iptable rule is missing".format(
                        cur_hostname_or_ip_with_port_and_protocol))
        return logging_server_status_dict

    def getHostNameOrIpAndPortAndProtocol(
            self, hostname_or_ip_address_with_port):
        res = hostname_or_ip_address_with_port.split('@')
        return res[0], res[1], res[2]

    def is_cert_expire(self, cert_pem):
        # Validate the certificate hasn't expired
        try:
            out = run_command(_GET_CERT_ENDDATE_CMD,
                input=cert_pem.encode(), shell=False) or ''
        except CalledProcessError as e:
            self._logger.warning(
                "Could not extract end date from the certificate")
            return None
        except Exception as e:
            self._logger.warning(
                "Error executing get extract certificate command: {}. "
                    "Error: {}".format(_GET_CERT_ENDDATE_CMD, e.stderr))
            return None
        self._logger.debug("output: {}".format(out))
        end_date = datetime.strptime(
            out.split("=")[1].strip(), _OPENSSL_DATE_FORMAT)
        self._logger.debug("Certificate's end date is {}".format(end_date))
        if end_date < datetime.now():
            self._logger.debug("Certificate has already "
                "expired on {}".format(end_date))
            return True
        return False

    def ipt_rule_exists(self, port, protocol, server=None,
            user=None, group=None, interface='eth0'):
        rule = []
        wait_option = ["-w", "5"]
        # Set the name of interface via which a packet
        # is going to be sent/receive
        if interface != "*":
            rule.extend(["-o", interface])
        if server:
            rule.extend(["-d", str(server)])
        if protocol == "TCP" or protocol == "TLS":
            rule.extend(["-p", "tcp", "-m", "tcp", "--dport", str(port),
                        "--tcp-flags", "FIN,SYN,RST,ACK", "SYN"])
        elif protocol == "UDP":
            rule.extend(["-p", "udp", "-m", "udp", "--dport", str(port)])
        else:
            self._logger.warning(
                "Invalid protocol specified {}".format(protocol))
            return None
        if user:
            rule.extend(["-m", "owner", "--uid-owner", user])
        if group:
            rule.extend(["-m", "owner", "--gid-owner", group])
        rule.extend(["-j", "ACCEPT"])
        # only check the OUTPUT CHAIN
        ipt_command = ["iptables"] + wait_option + ["-C", "OUTPUT"] + rule
        self._logger.debug("ipt_command: {}".format(ipt_command))
        return self.check_ipt_rule(ipt_command)

    def default_ipt_rule_exists(self, port):
        if port == _DEFAULT_TCP_PORT:
            return self.check_ipt_rule(_DEFAULT_TCP_IPT_RULE_CHECK_CMD)
        elif port == _DEFAULT_UDP_PORT:
            return self.check_ipt_rule(_DEFAULT_UDP_IPT_RULE_CHECK_CMD)
        return False

    def check_ipt_rule(self, ipt_command):
        try:
            out = run_command(ipt_command, shell=False) or ''
        except CalledProcessError as e:
            self._logger.debug(
                "Error executing iptables command: {}. Error: {}".format(
                     " ".join(ipt_command), e.stderr))
            return False
        except Exception as e:
            self._logger.warning(
                "Error executing iptables command: {}. Error: {}".format(
                     " ".join(ipt_command), e.stderr))
            return None
        return True