# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import sha.config

from ovsdbapp.backend.ovs_idl import idlutils

_logger = sha.config.get_logger(__name__)

METHOD_TRANSACTION = "transact"

TABLE_INTERFACE = "Interface"
TABLE_PORTS = "Port"

HYPERBUS = "hyperbus"

LINK_STATE_UP = "up"
LINK_STATE_DOWN = "down"

LACP_ACTIVE = "active"
LACP_PASSIVE = "passive"
LACP_OFF = "off"

UPLINK_DOWN_ALARM = "UPLINK_DOWN_ALARM"
LAG_MEMBER_DOWN_ALARM = "LAG_MEMBER_DOWN_ALARM"

_logger = sha.config.get_logger(__name__)


def set_uplink_down(status_changes):
    _logger.debug("Set uplink down alarm")
    status_changes[UPLINK_DOWN_ALARM] = True


def clear_uplink_down(status_changes):
    _logger.debug("Clear uplink down alarm")
    status_changes[UPLINK_DOWN_ALARM] = False


def set_lag_member_down(status_changes):
    _logger.debug("Set lag_member down alarm")
    status_changes[LAG_MEMBER_DOWN_ALARM] = True


def clear_lag_member_down(status_changes):
    _logger.debug("Clear lag_member down alarm")
    status_changes[LAG_MEMBER_DOWN_ALARM] = False


def is_interface_pnic(row):
    if row.type != "" \
            or row.external_ids.get("attached-mac",None) != None \
            or row.name == HYPERBUS:
        return False
    return True


def has_pnic_down(api, port):
    pnic_down = False
    idl = api.idl
    for col in port.interfaces:
        uuid = col.uuid
        interface = idlutils.row_by_value(idl, TABLE_INTERFACE, "uuid", uuid,
                                          None)
        if not is_interface_pnic(interface):
            continue

        if not interface:
            _logger.debug("Interface not found")
            continue
        if interface.link_state[0] == LINK_STATE_DOWN:
            pnic_down = True
            break
    return pnic_down


def has_lacp_down(api, port):
    lacp_down = False
    idl = api.idl
    for col in port.interfaces:
        uuid = col.uuid
        interface = idlutils.row_by_value(idl, TABLE_INTERFACE, "uuid", uuid,
                                          None)
        if not is_interface_pnic(interface):
            continue
        if not interface:
            _logger.debug("Interface not found")
            continue
        if interface.lacp_current[0] == False:
            lacp_down = True
            break
    return lacp_down


def all_lacp_down(api, port):
    lacp_down = False
    idl = api.idl
    for col in port.interfaces:
        uuid = col.uuid
        interface = idlutils.row_by_value(idl, TABLE_INTERFACE, "uuid", uuid,
                                          None)
        if not is_interface_pnic(interface):
            continue
        if not interface:
            _logger.debug("Interface not found")
            continue
        if interface.lacp_current[0] == True:
            lacp_down = False
            break
    return lacp_down


def is_lag_port(row):
    if len(row.lacp) <= 0:
        return False
    if row.lacp[0] == LACP_OFF:
        return False
    if row.lacp[0] == LACP_ACTIVE or row.lacp[0] == LACP_PASSIVE:
        return True
    return False


def has_uplink_down(api):
    uplink_down = False
    idl = api.idl
    port_table = idl.tables.get(TABLE_PORTS)
    for row in port_table.rows.values():
        if not is_lag_port(row):
            if has_pnic_down(api, row):
                uplink_down = True
                break
        else:
            if all_lacp_down(api, row):
                uplink_down = True
                break
    return uplink_down


def has_lag_member_down(api):
    lag_member_down = False
    idl = api.idl
    port_table = idl.tables.get(TABLE_PORTS)
    for row in port_table.rows.values():
        if is_lag_port(row):
            if has_lacp_down(api, row):
                lag_member_down = True
                break
    return lag_member_down


def has_interface_status_chane(new, old):
    if not (hasattr(new, "link_state") and hasattr(old, "link_state")):
        return True
    if not (hasattr(new, "lacp_current") and hasattr(old, "lacp_current")):
        return True
    if len(new.link_state) != len(old.link_state):
        _logger.info("link_state change")
        return True
    if len(new.link_state) > 0 and new.link_state[0] != old.link_state[0]:
        _logger.info("link_state change")
        return True
    if len(new.lacp_current) != len(old.lacp_current):
        _logger.info("lacp_current change")
        return True
    if len(new.lacp_current) > 0 and new.lacp_current[0] != old.lacp_current[
        0]:
        _logger.info("lacp_current change")
        return True
    return False


def check_overall_status(api, status_changes):
    if has_uplink_down(api):
        set_uplink_down(status_changes)
    else:
        clear_uplink_down(status_changes)

    if has_lag_member_down(api):
        set_lag_member_down(status_changes)
    else:
        clear_lag_member_down(status_changes)


def handle_notification(api, status_changes, event, row, updates=None):
    table_name = row._table.name
    if table_name != TABLE_INTERFACE:
        return
    if event == "delete":
        check_overall_status(api, status_changes)
        return
    if not is_interface_pnic(row):
        return
    if updates and not has_interface_status_chane(row, updates):
        _logger.debug("change not happend on interfaces")
        return
    if len(row.lacp_current) > 0 and row.lacp_current[0] is False:
        set_lag_member_down(status_changes)
        return
    if len(row.link_state) > 0 and row.link_state[0] == LINK_STATE_DOWN:
        set_uplink_down(status_changes)
        return
    check_overall_status(api, status_changes)
