# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import traceback

import sha.config
from sha.core.channel.provider.base import BaseProvider
from nsx_monitoring.clientlibrary.event_source import EventSource
from nsx_monitoring.clientlibrary.feature_event_type_mappings \
    import CNI_HEALTH_FEATURE_ID, NODE_AGENTS_HEALTH_FEATURE_ID, TRANSPORT_NODE_HEALTH_FEATURE_ID, \
    AUDIT_LOG_HEALTH_FEATURE_ID, COMMUNICATION_FEATURE_ID, AUDIT_LOG_HEALTH_REMOTE_LOGGING_SERVER_ERROR_EVENT_TYPE_ID, \
    INFRASTRUCTURE_SERVICE_FEATURE_ID, EDGE_HEALTH_FEATURE_ID, \
    COMMUNICATION_MANAGER_FQDN_REVERSE_LOOKUP_FAILURE_EVENT_TYPE_ID, \
    COMMUNICATION_MANAGER_FQDN_LOOKUP_FAILURE_EVENT_TYPE_ID


class AlarmProvider(BaseProvider):
    """The MPA Provider"""

    def __init__(self):
        self._name = "AlarmProvider"
        self._logger = sha.config.get_logger(__name__)

        config_callback = lambda event_type_config_map: print("received alarm callback", event_type_config_map)
        self.cni_health_source = EventSource(feature_id=CNI_HEALTH_FEATURE_ID, source_id="nsx_hyperbus_health")
        self.cni_health_source.set_config_callback(config_callback)
        self.node_agents_health_source = EventSource(feature_id=NODE_AGENTS_HEALTH_FEATURE_ID,
                                                source_id="nsx_node_agent_health")
        self.node_agents_health_source.set_config_callback(config_callback)
        self.tn_health_source = EventSource(feature_id=TRANSPORT_NODE_HEALTH_FEATURE_ID,
                                       source_id="nsx_transport_node_health")
        self.tn_health_source.set_config_callback(config_callback)
        self.audit_log_health_source = EventSource(feature_id=AUDIT_LOG_HEALTH_FEATURE_ID, source_id="nsx_audit_log_health")
        self.audit_log_health_source.set_config_callback(config_callback)
        self.communication_source = EventSource(feature_id=COMMUNICATION_FEATURE_ID,
                                                 source_id="nsx_communication")
        self.communication_source.set_config_callback(config_callback)
        self.infrastructure_service_source = EventSource(feature_id=INFRASTRUCTURE_SERVICE_FEATURE_ID,
                                                    source_id="nsx_infrastructure_service")
        self.infrastructure_service_source.set_config_callback(config_callback)
        self.cni_health_source.start()
        self.node_agents_health_source.start()
        self.tn_health_source.start()
        self.audit_log_health_source.start()
        self.communication_source.start()
        self.infrastructure_service_source.start()
        # Only for edge types
        if sha.config.node_type_is(sha.config.DIST_EDGE, sha.config.DIST_AUTONOMOUS_EDGE,
                                   sha.config.DIST_PUBLIC_CLOUD_GATEWAY):
            self.edge_health_source = EventSource(feature_id=EDGE_HEALTH_FEATURE_ID,
                                                  source_id="sha_edge_health")
            self.edge_health_source.set_config_callback(config_callback)
            self.edge_health_source.start()

    def sync(self, msg):
        try:
            self._logger.debug("{} is sending alarm msg:{} via event source instance".format(self.name, msg))
            if msg.feature == "cni_health":
                self.cni_health_source.update_event_instances(msg.status)
            elif msg.feature == "node_agents_health":
                self.node_agents_health_source.update_event_instances(msg.status)
            elif msg.feature == "transport_node_health":
                self.tn_health_source.update_event_instances(msg.status)
                self._logger.info('Delivered alarm event:{} to alarm framework successfully.'.format(msg))
            elif msg.feature == "audit_log_health":
                if msg.event_type == "remote_logging_server_error":
                    if len(msg.deleted_entity_list) > 0:
                        self._logger.warning("{} logging-server(s) are deleted".format(
                            len(msg.deleted_entity_list)))
                        self.audit_log_health_source.update_deleted_entities(
                            AUDIT_LOG_HEALTH_REMOTE_LOGGING_SERVER_ERROR_EVENT_TYPE_ID,
                            msg.deleted_entity_list)
                self.audit_log_health_source.update_event_instances(msg.status)
            elif msg.feature == "communication":
                self.communication_source.update_event_instances(msg.status)
                if msg.deleted_entity_list:
                    if msg.event_type == "manager_fqdn_reverse_lookup_failure":
                        self.communication_source.update_deleted_entities(
                            COMMUNICATION_MANAGER_FQDN_REVERSE_LOOKUP_FAILURE_EVENT_TYPE_ID,
                            msg.deleted_entity_list)
                    elif msg.event_type == "manager_fqdn_lookup_failure":
                        self.communication_source.update_deleted_entities(
                            COMMUNICATION_MANAGER_FQDN_LOOKUP_FAILURE_EVENT_TYPE_ID,
                            msg.deleted_entity_list)
            elif msg.feature == "infrastructure_service":
                self.infrastructure_service_source.update_event_instances(msg.status)
                self._logger.info('Delivered alarm event:{} to alarm framework successfully.'.format(msg))
            elif msg.feature == "edge_health":
                self.edge_health_source.update_event_instances(msg.status)
            else:
                self._logger.warning("unknown feature type is provided, no proper event source instance")
            self._logger.debug("{} finished sending alarm msg.".format(self.name, msg))
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("{} send alarm failed, msg:{}, Exception:{}, trace:{}"
                                 .format(self.name, msg, e, trace))

    def close(self):
        pass

    @property
    def name(self):
        return self._name


class AlarmMsg:
    def __init__(self, feature, status, event_type, deleted_entity_list):
        self._feature = feature
        self._status = status
        self._event_type = event_type
        self._deleted_entity_list = deleted_entity_list

    def __repr__(self):
        return 'AlarmMsg: feature:{}, status:{}, event_type:{}, deleted_entity_list:{}'\
            .format(self._feature, self._status, self._event_type, self._deleted_entity_list)

    @property
    def feature(self):
        return self._feature

    @property
    def status(self):
        return self._status

    @property
    def event_type(self):
        return self._event_type

    @property
    def deleted_entity_list(self):
        return self._deleted_entity_list
